#include "operator.hh"
#include "timers.hh"
#include "waves.h"


Operator::Operator() : outputlevel(0), target(0), source(0)
{
	envelopetimer = new timer;
	envpos1 = 4;
	envpos2 = 0;
}

Operator::~Operator()
{
	delete envelopetimer;
}

void Operator::Run(double secs)
{
	if(target)target->Run(secs);
	
	/* FIXME --- THIS IS LOTS OF FIXME */	
	unsigned ticks = envelopetimer->Advance(0.1, secs);
	
	for(; ticks > 0 && envpos1 < 4; --ticks)
	{
    	switch(envpos1)
    	{
    		case 0: /* attack */
    		{
    			double tmp = attack * envpos2;
    			
    			if(tmp < outputlevel)
    			{
    				effvol = tmp;
    				envpos2++;
    				break;
    			}
    			
    			effvol = outputlevel;
   				envpos1 = 1;
   				envpos2 = 0;
   				/* passthru */
   			}
   			case 1: /* decay */
   			{
   				double tmp = outputlevel - decay * envpos2 / 15.0;
   				if(tmp < 0)tmp = 0;
   				effvol = tmp;
   				
   				++envpos2;
   				
   				if(effvol > outputlevel*(1 - 1/(sustain+1)))
   					break;
   					
   				envpos1 = dosustain ? 2 : 3;
   				/* passthru */
   			}

   			case 2: /* sustain */
   			{
   				effvol = outputlevel*(1 - 1/(sustain+1));
   				return;
   			}
   			
   			case 3: /* release */
   			{
   				double tmp = outputlevel*(1 - 1/(sustain+1))
   				           - release * envpos2 / 15.0;
   				
   				++envpos2;
   				
   				if(tmp <= 0)
   				{
   					envpos1 = 4;
   					effvol = 0.0;
   					return;
	   			}
	   			
	   			effvol = tmp;
	   			break;
   			}
   		}
	}
}

void Operator::Modulate(Operator &o)
{
	o.source    = this;
	o.chaintype = mod_freq;
	target = &o;
}

void Operator::AddTo(Operator &o)
{
	o.source    = this;
	o.chaintype = mod_add;
	target = &o;
}

void Operator::Unlink(void)
{
	target = 0;
	source = 0;
}

double Operator::multiplyfreq(double freq) const
{
	switch(freqmul)
	{
		case 0: return freq/2;
		case 1: return freq;
		case 11: return freq * 10;
		case 13: return freq * 12;
		case 14: return freq * 15;
	}
	return freq * freqmul;
}

void Operator::Write20(unsigned char value)
{
	tremolo = value & 0x80;
	vibrato = value & 0x40;
	dosustain=value & 0x20;
	KSR =     value & 0x10;
	freqmul = value & 15;
}

void Operator::Write40(unsigned char value)
{
	KSL = value >> 6;
	
	unsigned attenuation = value & 63;
	unsigned power = 63 - attenuation;
	
	outputlevel = power * power / (63.0 * 63.0);
	
	effvol = outputlevel;
}

void Operator::Write60(unsigned char value)
{
	attack   = value >> 4;
	decay    = value & 15;
}
	
void Operator::Write80(unsigned char value)
{
	sustain  = value >> 4;
	release  = value & 15;
}

unsigned OperatorPortOffset(unsigned channel)
{
	unsigned ofs = 0x100 * (channel / 18);
	register unsigned chan18 = channel % 18;
	return ofs + (chan18%6) + 2 * (chan18/6);
}

unsigned PortOffsetOperator(unsigned offset)
{
	unsigned chan = 18 * (offset / 0x100);
	register unsigned low = offset & 0x1F;
	
	chan = low;
	
	if(low >= 0x08)chan -= 2;
	if(low >= 0x10)chan -= 2;
	
	if(chan >= 36)chan = 36-1;
	
	return chan;
}

static const double PII = 3.141592653589793;

double Operator::Produce(unsigned waveno, double freq,
                         double angle, double shift)
{
	if(effvol < 0.01)return 0.0;
	
	double result = waves[waveno&7] (multiplyfreq(freq) * angle + shift);
	result *= result;	
	result *= effvol;
	
	if(target)
	{
		if(chaintype == mod_add)
			result += target->Produce(waveno, freq, angle, 0.0);
		else
			result = target->Produce(waveno, freq, angle, PII * result);
	}
	return result;
}

void Operator::KeyOn()
{
	if(target)target->KeyOn();
	
	envpos1 = envpos2 = 0;
}

void Operator::KeyOff()
{
	if(target)target->KeyOff();
	envpos1 = 3;
	envpos2 = 0;
}
